/**
 * File for general "Load Balancing" logic.
 */


// User Import Headers.
#ifndef user_headers
    #define user_headers
    #include "load_balance_schemes.h"
    #include "load_balance_general.h"
    #include "structs.h"
    #include "terminal_commands.h"
#endif


/**
 * Entrypoint for logic to run "Asynchronous Round Robin" load scheme.
 */
void arr_run(thread_struct* thread_args_ptr) {
    int process_rank = thread_args_ptr->thread_num;

    if (process_rank == 0) {
        printf("Starting ARR load scheme.\n");
        printf("\n");
    }

    // Wait for all processes to synchronize.
    // Also sleep for a second to allow time for console output to process.
    MPI_Barrier(MPI_COMM_WORLD);
    sleep(1);

    // Handle based on process number.
    if (process_rank > 0) {
        // Child process. Handles all the work.
        arr_worker(thread_args_ptr);
    } else {
        // Main process. Handles communication to terminal.
        arr_main(thread_args_ptr);
    }

    // Wait for all processes to synchronize.
    // Also sleep for a second to allow time for console output to process.
    MPI_Barrier(MPI_COMM_WORLD);
    sleep(1);

    if (process_rank == 0) {
        printf("\n");
        printf("Finished ARR load scheme.\n");
        printf("\n");
    }
}


/**
 * Main processor logic for "Asynchronous Round Robin" load scheme.
 */
void arr_main(thread_struct* thread_args_ptr) {
    int index;
    int main_loop_bool = 1;
    int init_run = 1;
    int msg_status_flag;
    int msg_recieved = 0;
    int* recv_array;

    log("Starting arr_main.");

    // Initialize console display.
    for (index = 1; index < thread_args_ptr->total_processors; index++) {
        recv_array = calloc(3, sizeof(int));
        MPI_Recv(recv_array, 2, MPI_INT, index, tag_status_update, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        main_display_status(thread_args_ptr, recv_array[0], recv_array[1], init_run);
        free(recv_array);
    }

    // Continually grab messages from all worker processors until done.
    // For each loop, it checks all worker processes for a message about work being done.
    // If no messages are found for two consecutive loops in a row, then assumes all work is complete and terminates.
    index = 1;
    init_run = 0;
    while (main_loop_bool == 1) {
        msg_status_flag = 0;

        // Check if message present from respective process.
        MPI_Iprobe(index, tag_status_update, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
        while (msg_status_flag == 1) {
            msg_recieved = 1;

            // Message present. Retrieve and handle.
            recv_array = calloc(3, sizeof(int));
            MPI_Recv(recv_array, 2, MPI_INT, index, tag_status_update, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
            main_display_status(thread_args_ptr, recv_array[0], recv_array[1], init_run);
            free(recv_array);

            MPI_Iprobe(index, tag_status_update, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
        }

        // Increment for next thread.
        index = (index + 1) % thread_args_ptr->total_processors;

        // Check if past last processor.
        if (index == 0) {
            index += 1;

            // Check if messages were recieved this run.
            if (msg_recieved == 0) {
                // No messages recieved. Run one last loop to double check.
                msg_recieved = -1;
                sleep(2);
            } else if (msg_recieved == -1) {
                // No messages recieved for two loops in a row. Terminate.
                main_terminate_workers(thread_args_ptr);
                main_loop_bool = 0;
            } else {
                // One or more messages recieved this loop. Reset value.
                msg_recieved = 0;
            }

            // Sleep number of seconds provided by user terminal.
            // Meant to simulate "work", as it forces each worker process to block for this amount of time each load,
            // before continuing on. In a real program, the worker would be doing something useful in this time,
            // instead of just blocking.
            sleep(thread_args_ptr->seconds_per_load);
        }

    }

    log("Finished ARR main.");
}


/**
 * Worker processor logic for "Asynchronous Round Robin" load scheme.
 */
void arr_worker(thread_struct* thread_args_ptr) {
    int process_rank = thread_args_ptr->thread_num;
    int arr_counter = thread_args_ptr->thread_num;
    int main_loop_bool = 1;
    int msg_status_flag;
    int request_flag = -1;
    int index;
    int work_value = 0;
    int send_status = 1;

    log("Starting ARR worker.");

    // Initialize process "1" to have full loads. All others start with none.
    if (process_rank == 1) {
        thread_args_ptr->remaining_loads = thread_args_ptr->total_loads;
    } else {
        thread_args_ptr->remaining_loads = 0;
    }

    // Initialize output by immediately sending status to main process.
    worker_send_status(process_rank, thread_args_ptr->remaining_loads);

    // Main loop bool. Continues until main process says all work is done.
    while (main_loop_bool == 1) {

        // Loop all worker processes until load count message of "0" has been sent.
        while (thread_args_ptr->remaining_loads > 0) {

            // Send message for current load.
            worker_send_status(process_rank, thread_args_ptr->remaining_loads);

            // Check for message requests from other processes.
            msg_status_flag = 0;
            MPI_Iprobe(MPI_ANY_SOURCE, tag_work_request, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
            if (msg_status_flag == 1) {
                // At least one message pending. Handle before proceeding.
                worker_handle_request(thread_args_ptr);
            }

            // Decriment load number.
            thread_args_ptr->remaining_loads -= 1;
            send_status = 1;
        }

        // If we made it this far, then worker no longer has work.

        // Send notification message to main, to show value of 0.
        if (send_status == 1) {
            worker_send_status(process_rank, thread_args_ptr->remaining_loads);
            send_status = 0;
        }

        // Check for pending work requests.
        msg_status_flag = 0;
        MPI_Iprobe(MPI_ANY_SOURCE, tag_work_request, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
        if (msg_status_flag == 1) {
            // At least one message pending. Handle before proceeding.
            worker_handle_request(thread_args_ptr);
        }


        if (request_flag == -1) {
            // Update personal ARR counter.
            arr_counter = (arr_counter + 1) % thread_args_ptr->total_processors;
            while (arr_counter == thread_args_ptr->thread_num || arr_counter == 0) {
                arr_counter = (arr_counter + 1) % thread_args_ptr->total_processors;
            }

            // Beg fellow worker for work.
            worker_send_request(thread_args_ptr, arr_counter);
            request_flag = arr_counter;
        } else {
            msg_status_flag = 0;
            MPI_Iprobe(request_flag, tag_work_response, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
            if (msg_status_flag == 1) {
                MPI_Recv(&work_value, 1, MPI_INT, request_flag, tag_work_response, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

                // Update personal load value if non-reject message provided.
                if (work_value > 0) {
                    thread_args_ptr->remaining_loads = work_value;
                }

                // Reset flag to allow begging for work again in the future.
                request_flag = -1;
            }
        }

        // Check if program as a whole is done running.
        main_loop_bool = worker_check_status();

        // Check again for pending work requests.
        msg_status_flag = 0;
        MPI_Iprobe(MPI_ANY_SOURCE, tag_work_request, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
        if (msg_status_flag == 1) {
            // At least one message pending. Handle before proceeding.
            worker_handle_request(thread_args_ptr);
        }

        // Check status of overall work.
        if (main_loop_bool == 1) {
            // Sleep to simulate "doing work".
            sleep(1);
        }
    }

    // Work is complete. Stick around briefly to ensure all pending message requests are handled.
    for (index = 0; index < 5; index++) {
        worker_handle_request(thread_args_ptr);
        sleep(1);
    }

    log("Finished ARR worker.");
}


/**
 * Entrypoint for logic to run "Global Round Robin" load scheme.
 */
void grr_run(int total_processors) {
    printf("Starting GRR load scheme.");

    printf("Finished GRR load scheme.");
}


/**
 * Main processor logic for "Global Round Robin" load scheme.
 */
void grr_main(thread_struct* thread_args_ptr) {
    log("Starting GRR main.");

    log("Finished GRR main.");
}


/**
 * Worker processor logic for "Global Round Robin" load scheme.
 */
void grr_worker(thread_struct* thread_args_ptr) {
    log("Starting GRR worker.");

    log("Finished GRR worker.");
}


/**
 * Entrypoint for logic to run "Random Poling" load scheme.
 */
void rp_run(int total_processors) {
    log("Starting RP load scheme.");

    log("Finished RP load scheme.");
}


/**
 * Main processor logic for "Random Poling" load scheme.
 */
void rp_main(thread_struct* thread_args_ptr) {
    log("Starting RP main.");

    log("Finished RP main.");
}


/**
 * Worker processor logic for "Random Poling" load scheme.
 */
void rp_worker(thread_struct* thread_args_ptr) {
    log("Starting RP worker.");

    log("Finished RP worker.");
}



/**
 * Entrypoint for logic to run "Nearest Neighbor" load scheme.
 */
void nn_run(int total_processors) {
    log("Starting NN load scheme.");

    log("Finished NN load scheme.");
}


/**
 * Main processor logic for "Nearest Neighbor" load scheme.
 */
void nn_main(thread_struct* thread_args_ptr) {
    log("Starting NN main.");

    log("Finished NN main.");
}



/**
 * Worker processor logic for "Nearest Neighbor" load scheme.
 */
void nn_worker(thread_struct* thread_args_ptr) {
    log("Starting NN worker.");

    log("Finished NN worker.");
}
